package com.craigl.artrays;

import java.io.File;

public class ArtRays {
	
	// States of state machine
	private static final int STATE_SIGNON      = 0;
	private static final int STATE_SETUP_FILES = 1;
	private static final int STATE_RENDER_POV  = 2;
	private static final int STATE_DISPLAY_POV = 3;
	private static final int STATE_CLEANUP_POV = 4;
	
	public ArtRays() {
		
		// Instantiate random number generator\
		rng = new Randomize();
		
		// Instantiate the property manager
		pm = new PropertyManager();
		
		// Instantiate the POV Producer
		povp = new POVProducer(pm);
		
		// Instantiate the POV exec'er
		pove = new ExecPOV(pm);

		// Parameterize the instance from the property file
		width = pm.getWidth();
		height = pm.getHeight();
		
		pove.setWidth(width);
		pove.setHeight(height);
		pove.setAntiAliasing(pm.getAntialiasing());
		
		// Get the output dir path
		outputDir = pm.getOutputDir();
		
		// True stops state machine
		done = false;
		
		// Get save files flag
		saveFiles = pm.getSaving();

		// Get sound flag
		soundOn = pm.getSoundOn();
		
		// Get display flag
		displaying = pm.getDisplaying();

		// Get the vary size flag
		varySize = pm.getVarySize();
		
		// Set initial state of state machine
		if (displaying) {
			// Conditionally instantiate display class
			fsd = new FSDisplay();
			state = STATE_SIGNON;
		} else {
			state = STATE_SETUP_FILES;
		}
	}
	
	// Determine width of image to render
	private int getWidthToRender() {
		
		if (!varySize) {
			return width;
		}	else	{
			int index = rng.randomInt0toN(5);
			switch (index) {
			default:
			case 0:
				return 1024;
				
			case 1:
				return 832;
				
			case 2:
				return 640;
			}
		}
	}

	// Determine height of image to render
	private int getHeightToRender() {
		
		if (!varySize) {
			return height;
		}	else	{
			int index = rng.randomInt0toN(5);
			switch (index) {
			default:
			case 0:
				return 768;
				
			case 1:
				return 624;
				
			case 2:
				return 480;
			}
		}
	}

	// Run this code in an infinite loop
	private void run() {
		
		String povInputFileName = "";
		String povOutputFileName = "";
		String fullImageFileName = "";

		while (!done) {
			// Run the state machine
			switch (state) {

			case STATE_SIGNON:
				// Get the name of the signon image
				String fn = pm.getSignOnImage();
				
				// Display it full screen
				fsd.setDisplayImage(fn);
				
				// Set the next state
				state = STATE_SETUP_FILES;
				break;

			case STATE_SETUP_FILES:
				// Get the POV file to render
				povInputFileName = povp.getPovFile();

				// Parse out the name of the input file
				int index = povInputFileName.lastIndexOf('\\');
				String povfn = povInputFileName.substring(index + 1);
				String name = povfn.substring(0, povfn.length() - 4);
				povOutputFileName = outputDir + name + ".png";
				
				// Set the next state
				state = STATE_RENDER_POV;
				break;

			case STATE_RENDER_POV:
				
				// Set the width to render
				pove.setWidth(getWidthToRender());
				
				// Set the height to render
				pove.setHeight(getHeightToRender());
				
				// Set input file (POV file to render)
				pove.setInputFile(povInputFileName);
				
				// Set output file (POV PNG output file)
				pove.setOutputFile(povOutputFileName);
				
				// Render the image into the output file
				pove.execPOV();
				
				// Conditionally beep when image has been generated
				if (soundOn) {
					java.awt.Toolkit.getDefaultToolkit().beep();
				}

				// Set the next state
				if (displaying) {
					// If we are displaying generated images
					state = STATE_DISPLAY_POV;
				} else {
					// If we are not displaying generated images
					state = STATE_SETUP_FILES;
				}
				break;

			case STATE_DISPLAY_POV:
				
				// Create the full path/name of image file
				fullImageFileName = "\\artrays\\" + povOutputFileName;

				// Display the new image
				fsd.setDisplayImage(fullImageFileName);
				
				// Set the next state
				state = STATE_CLEANUP_POV;
				break;

			case STATE_CLEANUP_POV:
				// Determine if input file is a canned scene pov file or generated pov file
				boolean isSceneFile = povInputFileName.indexOf("scene") != -1;

				// Do conditional delete of input pov file
				if (!saveFiles && !isSceneFile) {
					// Delete the pov file
					new File(povInputFileName).delete();
				}

				// Do conditional delete of output image file
				if (!saveFiles) {
					// Delete the pov image file
					new File(fullImageFileName).delete();
				}

				// Set the next state
				state = STATE_SETUP_FILES;
				break;
			}
		}
	}

	// Art Rays Program entry point
	public static void main (String [] args) {
		
		ArtRays ar = new ArtRays();
		ar.run();
	}
	
	// Instance data
	private Randomize rng;
	private PropertyManager pm;
	private POVProducer povp;
	private ExecPOV pove;
	private FSDisplay fsd;

	private int width;
	private int height;
	private boolean done;
	private boolean saveFiles;
	private boolean soundOn;
	private boolean displaying;
	private boolean varySize;
	private int state;
	private String outputDir; 
}
