package com.craigl.artrays;

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.Date;

public class POVProducer {
	
	//  Our ray traced universe is 1000 x 1000 x 1000 units in size
	//  This results in the following bounds:
	//
	//    -500 <= x <=  500
	//    -500 <= y <=  500
	//       0 <= z <= 1000
	//
	// Dimensions of our universe
	private static final int UNIVERSE_SIDE = 1000;
	private static final int MIN_X = -UNIVERSE_SIDE/2;
	private static final int MAX_X =  UNIVERSE_SIDE/2;
	private static final int MIN_Y = -UNIVERSE_SIDE/2;
	private static final int MAX_Y =  UNIVERSE_SIDE/2;
	private static final int MIN_Z =  0;
	private static final int MAX_Z =  UNIVERSE_SIDE;
	
	// Various size radius
	private static final int RADIUS_SMALL  = 25;
	private static final int RADIUS_MEDIUM = 50;
	private static final int RADIUS_LARGE  = 100;
	private static final int RADIUS_XLARGE = 125;
	
	// Various size sides
	private static final int SIDE_SMALL  =  50;
	private static final int SIDE_MEDIUM = 100;
	private static final int SIDE_LARGE  = 200;
	private static final int SIDE_XLARGE = 250;

	private static final String EOL = "\r\n";
	private static final String INDENT = "   ";
	
	public POVProducer(PropertyManager pm) {
		
		// Instantiate the rng
		rng = new Randomize();
		
		// Save path to image map directory
		imageMapDir = pm.getImageMapDir();
		imDirFile = new File(imageMapDir);

		// Save path to input directory
		inputDir = pm.getInputDir();

		// Save path to scene directory
		sceneDir = pm.getSceneDir();
		sceneDirFile = new File(sceneDir);
		
		// Create a filter so that only .pov files from scene dir are returned
		povFilter = new FilenameFilter() {
	        public boolean accept(File dir, String name) {
	            return name.endsWith(".pov");
	        }
	    };
	    
		// Create a filter so that only image map files from imagemaps dir are returned
		imFilter = new FilenameFilter() {
	        public boolean accept(File dir, String name) {
	            return name.endsWith(".png");
	        }
	    };
	    
	    // Get scene parameter
	    useSceneEveryNImages = pm.useSceneEveryN();
	    modifyPaletteEveryNImages = pm.modifyPaletteEveryN();
	    
	    // Get max light sources
	    maxLights = pm.getMaxLights();

	    // Get shape constraints
	    minBoxes = pm.getMinBoxes();
	    maxBoxes = pm.getMaxBoxes();
	    minCones = pm.getMinCones();
	    maxCones = pm.getMaxCones();
	    minCyls = pm.getMinCyls();
	    maxCyls = pm.getMaxCyls();
	    minSpheres = pm.getMinSpheres();
	    maxSpheres = pm.getMaxSpheres();
	    
	    sceneCount = useSceneEveryNImages;
	    imageCount = modifyPaletteEveryNImages;
	    
	    pigmentRandomizer = pm.getPigmentRandomizer();
	}
	
	// Retrieve a random POV file from scene directory
	private String retrievePOVFile() {
		
		// Get all .pov files in scene directory
		String [] povFiles = sceneDirFile.list(povFilter);
	    
		// How many are there?
		int fileCount = povFiles.length;
	    
		// Pick a random index 0 <= randomIndex < fileCount
		int randomIndex = rng.randomInt0toN(fileCount);
	    
		// Return random scene file name
		return sceneDir + povFiles[randomIndex];	    
	}
	
	// Retrieve a random image map file from imagemaps directory
	private String retrieveImageMapFile() {
		
		// Get all .pov files in scene directory
		String [] imFiles = imDirFile.list(imFilter);
	    
		// How many are there?
		int fileCount = imFiles.length;
	    
		// Pick a random index 0 <= randomIndex < fileCount
		int randomIndex = rng.randomInt0toN(fileCount);
	    
		// Return random scene file name
		return imageMapDir + imFiles[randomIndex];	    
	}
	
	// Get a pov filename to render
	public String getPovFile() {
		
		// Is it time to throw in a scene image?
		if (--sceneCount == 0) {
			sceneCount = useSceneEveryNImages;
			
			// Return a scene POV file
			return retrievePOVFile();
		}	else	{
			if (--imageCount == 0) {
			    imageCount = modifyPaletteEveryNImages;
			    modifyPalette = true;
				paletteIndex = rng.randomInt0toN(8);

			}	else	{
				modifyPalette = false;
			}
			// Return a generated POV file
			return generatePOVFile();			
		}
	}
	
	// Generate a POV file
	private String generatePOVFile() {
		
		// Generate a filename to assign to the generated POV file
		String dateTimeFilename = generateDateTimeFilename();
		String fullFilename = inputDir + dateTimeFilename;
		
//		System.out.println("Full filename: " + fullFilename);
		
		// String buffer accumulates POV file elements
		StringBuffer sb = new StringBuffer();
		
		// Begin the generation of the POV file
		sb.append("// Autogenerated Art Rays POV file - Craig A. Lindley" + EOL + EOL);

		sb.append("#include \"colors.inc\"" + EOL + EOL);

		generateCamera(sb);
		sb.append(EOL);

		generateLightSources(sb);
		sb.append(EOL);

		// Define boxes in image
		int numberOfBoxes = rng.randomIntNtoM(minBoxes, maxBoxes);		
		for (int i = 0; i < numberOfBoxes; i++) {
			generateRandomBox(sb);
		}
		
		// Define cones in image
		int numberOfCones = rng.randomIntNtoM(minCones, maxCones);		
		for (int i = 0; i < numberOfCones; i++) {
			generateRandomCone(sb);
		}
		
		// Define cylinders in image
		int numberOfCylinders = rng.randomIntNtoM(minCyls, maxCyls);		
		for (int i = 0; i < numberOfCylinders; i++) {
			generateRandomCylinder(sb);
		}
		
		// Define spheres in image
		int numberOfSpheres = rng.randomIntNtoM(minSpheres, maxSpheres);		
		for (int i = 0; i < numberOfSpheres; i++) {
			generateRandomSphere(sb);
		}
		
		// Define floor in image
		generateRandomFloor(sb);
		
		// Define ceiling in image
		generateRandomCeiling(sb);
		
		// Define left wall in image
		generateRandomLeftWall(sb);
		
		// Define right wall in image
		generateRandomRightWall(sb);
		
		// Define back wall in image
		generateRandomBackWall(sb);
		
		// With image generated, write data to a file
	    try {
	    	FileWriter outFile = new FileWriter(fullFilename);
	    	PrintWriter out = new PrintWriter(outFile);
	    	out.println(sb.toString());
	    	out.close();
	    } catch (Exception e) {
	    	System.err.println(e.getMessage());
	    }
		return fullFilename;
	}
	
	// Generate a random color
	private void generateRandomColor(StringBuffer sb) {

		float red = rng.randomFloat();
		float green = rng.randomFloat();
		float blue = rng.randomFloat();
		
		// Are we modifying the palette of this image?
		if (modifyPalette) {
			// Calculate luminance of color
			float luminance = (float)(((30 * red) + (59 * green) + (11 * blue)) / 100.0);			
			
			switch (paletteIndex) {
			case 0: // 0 0 0
				red = green = blue = luminance;
				break;

			case 1: // 0 0 1
				blue = luminance;
				red = green = 0.0F;
				break;
				
			case 2: // 0 1 0
				green = luminance;
				red = blue = 0.0F;				
				break;
				
			case 3: // 0 1 1
				green = blue = luminance;
				red = 0.0F;				
				break;
				
			case 4: // 1 0 0
				red = luminance;
				green = blue = 0.0F;				
				break;
				
			case 5: // 1 0 1
				red = blue = luminance;
				green = 0.0F;				
				break;
				
			case 6: // 1 1 0
				red = green = luminance;
				blue = 0.0F;				
				break;
				
			case 7: // 1 1 1
				red = green = blue = luminance;
				break;
			}
		}
		
		sb.append("color rgb <" + red + "," + green + "," + blue + ">");
	}
	
	// Generate random point in our universe
	private void generateRandomPoint(StringBuffer sb, boolean toggleZ) {
		
		int x = rng.randomIntNtoM(MIN_X, MAX_X);
		int y = rng.randomIntNtoM(MIN_Y, MAX_Y);
		int z = rng.randomIntNtoM(MIN_Z, MAX_Z);
		
		z = (toggleZ) ? -1 * z : z;
		
		sb.append("<" + x + "," + y + "," + z + ">");
	}
	
	// Generate random camera location
	private void generateRandomCameraLocation(StringBuffer sb) {
		
		int x = rng.randomIntNtoM(MIN_X, MAX_X);
		int y = rng.randomIntNtoM(MIN_Y, MAX_Y);
		int z = rng.randomIntNtoM(-UNIVERSE_SIDE, MAX_Z/10);
		
		sb.append("location <" + x + "," + y + "," + z + ">");
	}
	
	private void generateCamera(StringBuffer sb) {
		
		sb.append("camera {" + EOL);
		sb.append(INDENT + "perspective" + EOL);
		sb.append(INDENT);
		generateRandomCameraLocation(sb);
		sb.append(EOL);
		sb.append(INDENT + "direction <0, 0, 1>" + EOL);
		sb.append(INDENT + "right 1.333 * x" + EOL);
		sb.append(INDENT + "up y" + EOL);
		sb.append(INDENT + "look_at <0, 0, 0>" + EOL);
		sb.append("}" + EOL);
	}
		
	private void generateLightSource(StringBuffer sb) {
		
		sb.append("light_source {" + EOL);
		sb.append(INDENT);
		generateRandomPoint(sb, true);
		sb.append(EOL + INDENT);
		generateRandomColor(sb);
		sb.append(EOL);		
		sb.append("}" + EOL + EOL);
	}
		
	private void generateLightSources(StringBuffer sb) {
		
		for (int l = 0; l < maxLights; l++) {
			// Generate a light source
			generateLightSource(sb);			
		}
	}
	
	private void generateRandomFinish(StringBuffer sb) {
		
		float ambient = rng.randomFloat();
		float diffuse = 1.0F - ambient;
		float brilliance = rng.randomFloatNtoM(1.0F, 300.0F);
		float phong = rng.randomFloat();
		float phongSize = rng.randomFloatNtoM(1.0F, 300.0F);
		float specular = rng.randomFloat();
		float roughness = rng.randomFloatNtoM(0.0005F, 1.0F);
		float metallic = rng.randomFloat();
		float crand = rng.randomFloatNtoM(0.01F, 0.5F);
		float reflection = rng.randomFloat();
		float irid = rng.randomFloatNtoM(0.0F, 0.5F);
		
		// Do we have a perfect mirror ?
		if (reflection >= 0.9F) {
			ambient = 0.0F;
			diffuse = 0.0F;
		}
		// Determine type of highlighting
		boolean usePhong = rng.randomBoolean();
		
		sb.append("finish {" + EOL);
		sb.append(INDENT + "ambient " + ambient + EOL);
		sb.append(INDENT + "diffuse " + diffuse + EOL);
		sb.append(INDENT + "brilliance " + brilliance + EOL);
		
		if (usePhong) {
			// Use phong
			sb.append(INDENT + "phong " + phong + EOL);
			sb.append(INDENT + "phong_size " + phongSize + EOL);

		}	else	{
			// Use specular
			sb.append(INDENT + "specular " + specular + EOL);
			sb.append(INDENT + "roughness " + roughness + EOL);
		}
		sb.append(INDENT + "metallic " + metallic + EOL);
		sb.append(INDENT + "crand " + crand + EOL);
		sb.append(INDENT + "reflection { " + reflection + " }" + EOL);
		sb.append(INDENT + "irid { " + irid + " }" + EOL);
		sb.append("}" + EOL);
	}
		
	// Generate a random pigment
	private void generateRandomPigment(StringBuffer sb) {
		
		sb.append("pigment {" + EOL);
		
		// Select a type of pigment
		// NOTE: index i chosen based upon property file value
		int i = rng.randomInt0toN(pigmentRandomizer);
		switch (i) {
		case 0:
		default:
			// Pigment is just a random color
			sb.append(INDENT);
			generateRandomColor(sb);
			sb.append(EOL);
			break;
			
		case 1:
			// Pigment is an image map			
			// Get a random image map path/filename
			// All files are PNG files so type not specified
			String imfn = retrieveImageMapFile();
			float scale = rng.randomFloatNtoM(50.0F, 400.0F);
			boolean once = rng.randomBoolean();
			
			sb.append(INDENT + "image_map {png \"" + imfn + "\"");
			sb.append((once) ? " once" : "");
			sb.append(" }" + EOL);
			sb.append(INDENT + "scale " + scale + EOL);
			sb.append(INDENT + "translate <-320, -240, 0>" + EOL);
			break;
			
		case 2:
			// Pigment is brick pattern
			sb.append(INDENT + "brick ");
			generateRandomColor(sb);
			sb.append(", ");
			generateRandomColor(sb);
			break;

		case 3:
			// Pigment is checker pattern
			sb.append(INDENT + "checker ");
			generateRandomColor(sb);
			sb.append(", ");
			generateRandomColor(sb);
			break;

		case 4:
			// Pigment is hex pattern
			sb.append(INDENT + "hexagon ");
			generateRandomColor(sb);
			sb.append(", ");
			generateRandomColor(sb);
			sb.append(", ");
			generateRandomColor(sb);
			break;
		}
		sb.append(EOL + "}" + EOL);
	}

	// Generate a random normal
	private void generateRandomNormal(StringBuffer sb) {

		sb.append("normal {" + EOL);

		float size = rng.randomFloat();
		float scale = rng.randomFloatNtoM(1.0F, 30.0F);

		// Select a type of normal
		int i = rng.randomInt0toN(6);
		switch (i) {
		case 0:
		default:
			// Agate as normal
			sb.append(INDENT + "agate " + size + EOL);
			break;

		case 1:
			// Brick as normal
			sb.append(INDENT + "brick " + size + EOL);
			break;

		case 2:
			// Bumps as normal
			sb.append(INDENT + "bumps " + size + EOL);
			break;

		case 3:
			// Checker as normal
			sb.append(INDENT + "checker " + size + EOL);
			break;

		case 4:
			// Crackle as normal
			sb.append(INDENT + "crackle " + size + EOL);
			break;

		case 5:
			// Dents as normal
			sb.append(INDENT + "dents " + size + EOL);
			break;
		}
		sb.append(INDENT + "scale " + scale + EOL);
		sb.append("}" + EOL);
	}
	
	// Generate a random texture
	private void generateRandomTexture(StringBuffer sb) {
		
		// Randomly incorporate normal into texture
		boolean useNormal = rng.randomBoolean();
		
		sb.append("texture {" + EOL);
		generateRandomPigment(sb);
		if (useNormal) {
			generateRandomNormal(sb);
		}
		generateRandomFinish(sb);
		sb.append("}" + EOL);
	}
	
	// Generate a random box
	private void generateRandomBox(StringBuffer sb) {
		
		// Generate random dimensions
		int width = randomSide();
		int height = randomSide();
		int depth = randomSide();
	
		// Generate random position
		float x = randomXFloat();
		float y = randomYFloat();
		float z = randomZFloat();
		
		float x1 = x + width;
		float y1 = y + height;
		float z1 = z + depth;
		
		sb.append("box {" + EOL);
		sb.append("<" + x + "," + y + "," + z + ">, ");
		sb.append("<" + x1 + "," + y1 + "," + z1 + ">" + EOL);
		generateRandomTexture(sb);
		sb.append("}" + EOL);
	}

	// Generate a random cone
	private void generateRandomCone(StringBuffer sb) {
		
		boolean pointed = rng.randomBoolean();
		boolean open = rng.randomBoolean();
		
		// Generate random base point
		float x = randomXFloat();
		float y = randomYFloat();
		float z = randomZFloat();
		
		float x1 = x;
		float y1 = y;
		float z1 = z;
		
		int side = randomSide();
		int capRadius = randomRadius();
		if (pointed) {
			capRadius = 0;
		}
		int baseRadius = randomRadius();
		
		// Select a axis for cone
		int i = rng.randomInt0toN(3);
		switch (i) {
		case 0:
			// Along x axis
			x1 += side;
			break;
			
		case 1:
			// Along y axis
			y1 += side;
			break;
			
		case 2:
			// Along z axis
			z1 += side;
			break;
		}
		sb.append("cone {" + EOL);
		sb.append("<" + x + "," + y + "," + z + ">, " + baseRadius + ", ");
		sb.append("<" + x1 + "," + y1 + "," + z1 + ">, " + capRadius + EOL);
		if (open) {
			sb.append("open" + EOL);
		}
		generateRandomTexture(sb);
		sb.append("}" + EOL);
	}

	// Generate a random cylinder
	private void generateRandomCylinder(StringBuffer sb) {
		
		boolean open = rng.randomBoolean();
		
		// Generate random base point
		float x = randomXFloat();
		float y = randomYFloat();
		float z = randomZFloat();
		
		float x1 = x;
		float y1 = y;
		float z1 = z;
		
		int side = randomSide();
		int radius = randomRadius();
		
		// Select a axis for cylinder
		int i = rng.randomInt0toN(3);
		switch (i) {
		case 0:
			// Along x axis
			x1 += side;
			break;
			
		case 1:
			// Along y axis
			y1 += side;
			break;
			
		case 2:
			// Along z axis
			z1 += side;
			break;
		}
		sb.append("cylinder {" + EOL);
		sb.append("<" + x + "," + y + "," + z + ">, ");
		sb.append("<" + x1 + "," + y1 + "," + z1 + ">, " + radius + EOL);
		if (open) {
			sb.append("open" + EOL);
		}
		generateRandomTexture(sb);
		sb.append("}" + EOL);
	}

	// Generate a random sphere
	private void generateRandomSphere(StringBuffer sb) {
		
		// Generate random base point
		float x = randomXFloat();
		float y = randomYFloat();
		float z = randomZFloat();
		
		int radius = randomRadius();
		
		sb.append("sphere {" + EOL);
		sb.append("<" + x + "," + y + "," + z + ">, " + radius + EOL);
		generateRandomTexture(sb);
		sb.append("}" + EOL);
	}

	// Generate a random floor
	private void generateRandomFloor(StringBuffer sb) {
		
		boolean hasFloor = rng.randomBoolean();
		if (hasFloor) {
			sb.append("// Floor" + EOL);
			float distance = -rng.randomFloatNtoM(50.0F, MAX_Y);
			sb.append("plane {" + EOL);
			sb.append("y, " + distance + EOL);
			generateRandomTexture(sb);
			sb.append("}" + EOL);
		}
	}

	// Generate a random ceiling
	private void generateRandomCeiling(StringBuffer sb) {
		
		boolean hasCeiling = rng.randomBoolean();
		if (hasCeiling) {
			sb.append("// Ceiling" + EOL);
			float distance = rng.randomFloatNtoM(100.0F, MAX_Y);
			sb.append("plane {" + EOL);
			sb.append("y, " + distance + EOL);
			generateRandomTexture(sb);
			sb.append("}" + EOL);
		}
	}

	// Generate a random left wall
	private void generateRandomLeftWall(StringBuffer sb) {
		
		boolean hasLeftWall = rng.randomBoolean();
		if (hasLeftWall) {
			sb.append("// Left wall" + EOL);
			float distance = -rng.randomFloatNtoM(100.0F, MAX_X);
			sb.append("plane {" + EOL);
			sb.append("x, " + distance + EOL);
			generateRandomTexture(sb);
			sb.append("}" + EOL);
		}
	}

	// Generate a random right wall
	private void generateRandomRightWall(StringBuffer sb) {
		
		boolean hasRightWall = rng.randomBoolean();
		if (hasRightWall) {
			sb.append("// Right wall" + EOL);
			float distance = rng.randomFloatNtoM(100.0F, MAX_X);
			sb.append("plane {" + EOL);
			sb.append("x, " + distance + EOL);
			generateRandomTexture(sb);
			sb.append("}" + EOL);
		}
	}

	// Generate a random back wall
	private void generateRandomBackWall(StringBuffer sb) {
		
		boolean hasBackWall = rng.randomBoolean();
		if (hasBackWall) {
			sb.append("// Back wall" + EOL);
			float distance = rng.randomFloatNtoM(200.0F, MAX_Z);
			sb.append("plane {" + EOL);
			sb.append("z, " + distance + EOL);
			generateRandomTexture(sb);
			sb.append("}" + EOL);
		}
	}

	// Return a random X value contained in our universe
	private float randomXFloat() {
		
		return rng.randomFloatNtoM(MIN_X, MAX_X);
	}

	// Return a random Y value contained in our universe
	private float randomYFloat() {
		
		return rng.randomFloatNtoM(MIN_Y, MAX_Y);
	}

	// Return a random Z value contained in our universe
	private float randomZFloat() {
		
		return rng.randomFloatNtoM(MIN_Z, MAX_Z);
	}

	private int randomRadius() {
		
		int i = rng.randomInt0toN(4);
		switch (i) {
		case 0:
			return RADIUS_SMALL;
		case 1:
			return RADIUS_MEDIUM;
		case 2:
			return RADIUS_LARGE;
		case 3:
			return RADIUS_XLARGE;
		default:
			return RADIUS_MEDIUM;
		}
	}
		
	private int randomSide() {
		
		int i = rng.randomInt0toN(4);
		switch (i) {
		case 0:
			return SIDE_SMALL;
		case 1:
			return SIDE_MEDIUM;
		case 2:
			return SIDE_LARGE;
		case 3:
			return SIDE_XLARGE;
		default:
			return SIDE_MEDIUM;
		}
	}
	
	// Generate a unique filename. Unique to the second
	private String generateDateTimeFilename() {
		
		Date now = new Date();
		SimpleDateFormat formatter = new SimpleDateFormat("MMddyy-HHmmss");
		return "F" + formatter.format(now) + ".pov";
	}

	// Test entry point
	public static void main(String [] args) {
		PropertyManager pm = new PropertyManager();		
		POVProducer povp = new POVProducer(pm);		
		povp.generatePOVFile();
	}
	
	// Instance data
	private Randomize rng;
	private FilenameFilter povFilter;
	private FilenameFilter imFilter;
	private String imageMapDir;
	private String inputDir;
	private String sceneDir;
	private File imDirFile;
	private File sceneDirFile;
	
	private int useSceneEveryNImages;
	private int modifyPaletteEveryNImages;
	
	private int maxLights;
	private int minBoxes;
	private int maxBoxes;
	private int minCones;
	private int maxCones;
	private int minCyls;
	private int maxCyls;
	private int minSpheres;
	private int maxSpheres;	
	
	private int imageCount;
	private int sceneCount;
	private boolean modifyPalette;
	private int paletteIndex;
	private int pigmentRandomizer;
}
